Huawei Technologies Co., Ltd
        
Huawei Serverless Sandbox User Guide
 

All Rights Reserved
Table of Contents
1 Sandbox Introduction  4
1.1 Feature Summary 4
1.2 Definitions and Abbreviations   4
1.3 References  5
2 Installation and Configuration    5
2.1 Host OS support 5
2.2 Runtimes    5
2.3 HSS Download    6
2.3.1 HSS Package   6
2.3.2 Language Runtime Packages 6
2.4 Binaries    6
2.5 Configuration File  6
2.5.1 Access Key and Secret Key 6
2.5.2 Region Name, Project Name and Project Id  7
2.5.3 DMS Endpoint  7
2.5.4 IAM Endpoint  8
3 Getting Started   9
3.1 Getting Help    9
3.1.1 General Help  9
3.1.2 Command Help  9
3.2 Invoking a Function 9
3.2.1 Usage Examples    10
3.3 Generating Local Events 10
3.3.1 Service Options   11
3.3.2 Event Details 11
3.4 Sourcing Events from Huawei Cloud Services  14
3.5 Validating a HSAM Template  15
4 Building a Simple Application Using Sandbox   16
4.1 Examples    16
4.1.1 Python    16
4.1.2 Go    16
4.2 Local Logging   17
4.3 Using an Environment variables  17
4.3.1 HSAM Template 17
4.3.2 Environment Variable file 18
4.3.3 Shell Environment 18
4.3.4 Combination of Shell and Environment Variable file    18
4.4 Debugging   19
4.4.1 Debugging Python Functions    19
4.4.2 Debugging Go Functions    19

1  Sandbox Introduction
Huawei Serverless Sandbox (HSS) is used to develop and test serverless applications on your local host machine before deployment to the Huawei Cloud.
Serverless applications are applications composed of Function Stages triggered by events.  Huawei Function Stages may be triggered by events such as object uploads to Object Storage Service (OBS), or messages received at a Distributed Message Service (DMS) queue, or the SMN service.
Huawei Serverless Sandbox (HSS) can be used to test these Function Stages locally, validate a Huawei Serverless Application Model (HSAM) template, and generate sample payloads locally for various event sources.
HSS has a cloudEvent command in which events from a Huawei cloud service can be routed directly to the local environment to invoke the local Function Stage.

1.1  Feature Summary
•   Develop and test Function Stages on your local host machine with HSS.
•   Invoke Function Stages using local event files. 
•   Invoke Function Stages from Huawei cloud services such as OBS, SMN, and DMS.
•   Validate Function Stage HSAM templates.
•   Debug Function Stages.
1.2  Definitions and Abbreviations
•   AK – Access Key
•   AOS – Application Orchestration Service
•   CES – Cloud Eye Service
•   CTS – Cloud Trace Service
•   DES – Data Express Service
•   DMS – Distributed Message Service
•   FSS – Function Stage Service
•   HSAM – Huawei Serverless Application Model
•   HSS – Huawei Serverless Sandbox 
•   OBS – Object Storage Service/ Object Block Storage
•   SMN – Simple Message Notification
•   SK – Secret Key
•   YAML – YAML Ain’t Markup Language
1.3  References
•   Huawei Function Stage Service (FSS) API Reference
•   Huawei Distributed Messaging System http://support.huaweicloud.com/en-us/dms_dld/index.html
•   Huawei Simple Message Notification http://support.huaweicloud.com/en-us/smn_gls/index.html
•   Huawei Object Storage Service http://support.huaweicloud.com/en-us/obs_gls/index.html
•   Huawei CodeHub http://www.huaweicloud.com/en-us/product/codehub.html
•   YAML.org http://yaml.org/
•   YAML Validator http://www.yamllint.com/
•   JSON Validator https://jsonlint.com/
2  Installation and Configuration
Running HSS to develop Function Stages on a local host machine requires Docker to be installed and running on the host machine. HSS will use the DOCKER_HOST environment variable to contact the Docker daemon.
To verify that Docker is working, you can run Docker commands from the CLI, for example: docker ps.
2.1  Host OS support
Linux is supported initially. Windows and Mac support will be added.
2.2  Runtimes
Supported runtimes include:
•   Python 2.7
•   Python 3.6
•   Go 1.8.3
2.3  HSS Download
2.3.1  HSS Package
The HSS package (hss.<RELEASE_VERSION>.package.tar.gz) can be downloaded from TBD. [HQ to add details]. This package includes the hss executable, config.yaml, README.md and examples.
Create a directory and untar the package
mkdir hss
tar xvfz hss.<RELEASE_VERSION>.package.tar.gz

2.3.2  Language Runtime Packages
The language runtime packages can be individually downloaded from TBD. [HQ to add details].  In the current release, three runtime packages are provided. They are
sandbox-py2.7-runtime.tar.gz, sandbox-py3.6-runtime.tar.gz and sandbox-go1.8-runtime.tar.gz for Python2.7, Python3.6 and Golang1.8 respectively.
Download either one (or all) the runtimes as needed.  
Create a directory and untar the package.
mkdir runtimes
tar xvfz sandbox-py2.7-runtime.tar.gz
This will create a file called “cff-sandbox-py2.7-runtime.tar” which is the docker runtime image for Python2.7.
Load this docker image using
docker load –i cff-sandbox-py2.7-runtime.tar
root@xx:/opt/hss/runtimes# docker load -i cff-sandbox-py2.7-runtime.tar 
1067b21d8ccd: Loading layer [==================================================>]  205.1MB/205.1MB
b1ae11fa8e0c: Loading layer [==================================================>]  76.95MB/76.95MB
a45dd5e27cea: Loading layer [==================================================>]   42.5kB/42.5kB
ca0e4e286f7e: Loading layer [==================================================>]  6.656kB/6.656kB
Loaded image: sandbox:py2.7-runtime
root@xx:/opt/hss/runtimes# docker images
REPOSITORY          TAG                 IMAGE ID            CREATED             SIZE
sandbox             py2.7-runtime       97d6d0ca8557        14 minutes ago      275MB

2.4  Binaries
Only binary packages will be released.
2.5  Configuration File
HSS uses a configuration file config.yaml which specifies the following:
•   Access Key and Secret Key
•   Region, Project Name and Identifier
•   IAM and DMS service endpoints
An example of a config.yaml is shown below.

  AccessKey: "12495NTRLMERTMS3DY5H"
  SecretKey: "CCkhEDdIA3q4wKTiUWmXtrwKCPz989fOvd9CVulG1"
  RegionName: "southchina"
  ProjectName: "southchina"
  ProjectId: "127eb1454f3d4d529813cb8523c7dc45"
  DMSEndpoint: "dms.southchina.myhuaweicloud.com"
  IAMEndpoint: "iam.cn-north-1.myhwclouds.com"

You must configure config.yaml with your own values which can be obtained from the Huawei Cloud portal as described below.
2.5.1  Access Key and Secret Key
Log in to your Huawei cloud account at http://www.huaweicloud.com/. Under your user name, select ‘My Credential’ to display your Access Keys. Edit your config.yaml file and add the Access Key and your Secret Key.

 
2.5.2  Region Name, Project Name and Project Id
Log in to your Huawei cloud account. Under your user name, select ‘My Credential’ to display your Project List. Edit your config.yaml file and add the Region Name, Project Name and Project Id.
 
2.5.3  DMS Endpoint
The DMS endpoint is used to configure the DMS queue for the cloudEvent command and can be obtained from this location: http://developer.huaweicloud.com/endpoint.
 
2.5.4  IAM Endpoint
The IAM endpoint is used for authentication and can be obtained from this location: http://developer.huaweicloud.com/endpoint.
 
3  Getting Started
3.1  Getting Help
3.1.1  General Help
To get help type the following:
  $ hss --help

3.1.2  Command Help
To get help for a specific command type the following:

  $ hss <command> --help

3.2  Invoking a Function
Use the following command to invoke a Function Stage.

  $ hss invoke [options] [<FunctionStage>]

Options:
--templateFile, -t  Path to the template file   Required. Default: hss.[yaml|yml]
--debugPort, -d Debug port number   Optional
--eventFile, -e JSON event file for simulated events    Optional
--logFile, -l   Log file name   Optional
--envFile, -n   Environment variable file name  Optional
--verbose, -v   Provide verbose debug information   Optional

Arguments:
FunctionStage   Function stage name.
If the template file contains more than one Function Stage, the name of the Function Stage must be specified, otherwise the command will terminate with an error.

3.2.1  Usage Examples
HSS will use the default HSAM template file template.yaml in the current working directory if a template file is not specified. If the template file has only a single Function Stage, it will invoke that Function Stage. A Function Stage can be invoked using a event file and the output will be sent to stdout.

  $ hss invoke –e event.json

A Function Stage can be invoked be piping a input event from stdin:

  $ echo '{"keyword1": "value2" }' | hss invoke MyFuncStage1

If hss invoke has no event file and no stdin value, then it will wait for you to enter an event string. You can enter ‘control-D’ to invoke the Function stage with no input. 
A Function Stage can be invoked using an explicit template file:

  $ hss invoke -e event.json -t template.yml  MyFuncStage1

3.3  Generating Local Events
Use the following command to generate mock JSON event records that can be used when invoking a Function Stage.

  $ hss localEvent <service> [service-options]

The output from the hss localEvent command can be output to a local file, and then this file can be used in an hss invoke command as shown below.

  $ hss localEvent obs --bucket bucket1 > event.json
  $ hss invoke –e events.json <FunctionStage>

Alternatively, the output from this command can be piped directly to an hss invoke command as shown below.

  $ hss localEvent obs --bucket bucket2 | hss invoke <FunctionStage>

If an event file is specified in addition to using piped input from stdin it will take precedence over the input from stdin.
3.3.1  Service Options
This command can be used to generate typical events from Huawei cloud services such as OBS, SMN, and DMS. <service> can be one of the following
obs     - generates a Huawei OBS event
smn     - generates a Huawei SMN event
dms     - generates a Huawei DMS event

OBS options:
--region, -r        OBS region  Optional. Def: cn-north-1
--bucket, -b        OBS bucket name Optional. Def: example-bucket
--key, -k           OBS key name    Optional. Def: example/key

SMN options:
--message, -m       SMN message Optional. Def: <default message>
--subject, -s       SMN subject Optional. Def: <default subject>
--topic, -t             SMN topic   Optional. Def: <default topic>

DMS options:
--region, -r        DMS region  Optional. Def: cn-north-1
--queueId, -qid     DMS queue id    Optional. Def: <default QID>
--consumerGroupId, -gid     DMS consumer group id   Optional. Def: <default CGID>
--messages, -m      DMS message Optional. Def: <default messages>


3.3.2  Event Details
[HQ: Either move this section to an appendix or remove it.]
DMS Record
  {
    "QueueID": "39907254-5321-4b90-bad2-f82455ff3ead",
    "Region": "cn-north-1",
    "Messages": [
        {
            "Body": "this is body message",
            "Attributes": {
                "weight": "80",
                "size": "100"
            }
        }
    ],
    "EventType": "MessageCreated",
    "ConsumerGroupID": "g-06c2bff5-67db-4055-ba4e-a31bb09c3485"
  }
OBS Record
{
    "Records": [
        {
            "eventVersion": "2.0",
            "eventTime": "2018-01-09T07:50:50.028Z",
            "requestParameters": {
                "sourceIPAddress": "103.218.216.125"
            },
            "s3": {
                "configurationId": "UK1DGFPYUKUZFHNQ00000160CC0B471D101ED30CE24DF4DB",
                "object": {
                    "eTag": "9d377b10ce778c4938b3c7e2c63a229a",
                    "sequencer": "00000000160D9E681484D6B4C0000000",
                    "key": "dogs.jpg",
                    "size": 777835
                },
                "bucket": {
                    "arn": "arn:aws:s3:::syj-input2",
                    "name": "syj-input2",
                    "ownerIdentity": {
                        "PrincipalId": "0ed1b73473f24134a478962e631651eb"
                    }
                },
                "s3SchemaVersion": "1.0"
            },
            "responseElements": {
                "x-amz-id-2": "91o45L3piWz75nQf7fl0dSQDgkaXVR2jZO3NZArg1H/wtNCeEjr43sdVsXJNYUjo",
                "x-amz-request-id": "100D00000160D9E67FACCEC517CCA7EE"
            },
            "awsRegion": "cn-north-1",
            "eventName": "ObjectCreated:Post",
            "userIdentity": {
                "principalId": "9bf43789b1ff4b679040f35cc4f0dc05"
            },
            "eventSource": "aws:s3"
        }
    ]
}
SMN Record
  {
    "record": [
        {
            "event_version": "1.0",
            "smn": {
                "topic_urn": "urn:smn:cn-north-1:0162c0f220284698b77a3d264376343a:test3",
                "timestamp": "2018-01-09T07:11:40Z",
                "message_attributes": null,
                "message": "this is smn message content",
                "type": "notification",
                "message_id": "a51671f77d4a479cacb09e2cd591a983",
                "subject": "this is smn message subject"
            },
            "event_subscription_urn":
                "urn:fss:cn-north-1:0162c0f220284698b77a3d264376343a:function:default:
                      read-smn-message:latest",
            "event_source": "smn"
        }
    ]
  }

3.4  Sourcing Events from Huawei Cloud Services 
The HSS cloudEvent command will source events from a Huawei cloud service such as OBS. This is done using a DMS queue. You must configure the Huawei cloud service to send events to a DMS queue. HSS will poll for events from the DMS queue and then send the event to stdout.
 
The syntax for this command is as follows.

  $ hss cloudEvent [options]

The event from stdout of the hss cloudEvent command can be output to a local file, and then this file can be used in an hss invoke command as shown below.

  $ hss cloudEvent -qid <qid> -gid <gid> > event.json
  $ hss invoke –e event.json <FunctionStage>

Alternatively, the stdout from this command can be piped directly to an hss invoke command as shown below.

  $ hss cloudEvent -qid <qid> -gid <gid>  | hss invoke <FunctionStage>

Options:
--interval, -i          Time interval to poll DMS queue.
Range: 1-10 seconds. Default = 2 seconds.   Optional
--queueId, -qid DMS queue id    Required
--consumerGroupId, -gid DMS consumer group id   Required

You must configure Huawei cloud authentication and configure the Huawei cloud service to route the events to the DMS queue.
When the hss cloudEvent command is invoked, HSS starts polling the DMS queue in the Huawei cloud every interval seconds where interval is between 1 and 10 seconds. When an event is found in the DMS queue, it is retrieved and sent to stdout. 

3.5  Validating a HSAM Template
Use the following command to validate a Function Stage HSAM template file. The name of the template file is specified using the –t or –-templateFile option. If a template file is not specified in the command, a default template file template.yaml in the current working directory will be validated. This command syntax is as follows:

  $ hss validate
  $ hss validate [–t (--template) <template-file>]

Usage examples:
To validate template.yaml in the current working directory:

  $ hss validate

To validate a file at /root/templates/test.yaml:

  % hss validate –t /root/templates/test.yaml

4  Building a Simple Application Using Sandbox
4.1  Examples
Examples of Function Stages for different languages are shown below.
4.1.1  Python
An example of the structure for a Python handler function is shown below. 
import json
def printhello():
    print 'Hello world!'

def handler (event, context):
    output = json.dumps(event)
    printhello()
return output
4.1.2  Go
An example ot the structure for a Go function is shown below. 
package main

import (
  "fmt"
  "go-runtime/go-api/context"
)
// Handler function
func Handler(event []byte, ctx context.RuntimeContext) (interface{}, error) {
  fmt.Println("\nIn serverless function...")
  ctxlog := ctx.GetLogger()
  ctxlog.Logf("RemainingTimeInMillis:%v",
              ctx.GetRemainingTimeInMilliSeconds())
  ctxlog.Logf("EventData :%v", string(event))
  return string(event), nil
}
4.2  Local Logging
The invoke command allow you to pipe logs from the invocation of the Function Stage into a file. This will be useful if you are running automated tests against HSS and want to capture logs for analysis. For example:
  $ hss invoke --logFile ./output.log

4.3  Using an Environment variables
If your Function Stage uses environment variables, values for them can be passed to the Docker container from three possible sources. Each environment variable will be selected from these three sources in the following precedence order:
1.  Environment variable file specified with the --envFile option.
2.  Shell's environment variable.
3.  Defined in Function Properties of template file.
In other words, an environment variable that is defined in a higher precedence source will used instead of the same environment variable is defined in a lower precedence source.
4.3.1  HSAM Template
For example, consider the HSAM template with two functions MyFunc1 and MyFunc2.
SAMversion: 'v1' 
Resources:
  MyFunc1:
    Type: HEC::Serverless::Function
    Properties:
      Handler: gotest.handler
      Runtime: go1.8.3
      MemorySize: 1024
      Timeout: 30
      Environment:
        Variables:
          TABLE_NAME: my-table
  MyFunc2:
    Type: HEC::Serverless::Function
    Properties:
      Handler: pytest.handler
      Runtime: python2.7
      Environment:
        Variables:
          STAGE: prod
          TABLE_NAME: my-table

4.3.2  Environment Variable file
Use the --envFile argument of invoke command to provide a YAML file that contains values for environment variables defined in your function. For example:
   NAME: John
   LAST_NAME: Smith

For example:
  $ hss invoke --envFile env.yaml

4.3.3  Shell Environment
Environment variables defined in your shell's environment will be passed to the Docker container, if they map to an environment variable in your Function Stage. Shell variables are globally applicable to all functions. For example, if two functions use an environment variable called TABLE_NAME, then the value for TABLE_NAME provided through the shell's environment will be available to both functions.
Following command will make value of TABLE_NAME available to both MyFunction1 and MyFunction2.
  $ TABLE_NAME=mytable hss invoke
4.3.4  Combination of Shell and Environment Variable file
For greater control, you can use a combination of shell environment variables and an environment variable file. If a variable is defined in both places, the one from the file will override the environment variable defined in the shell.
4.4  Debugging
The command hss invoke supports local debugging of your Function Stages. To run HSS with debugging support enabled, specify --debugPort or -d on the command line. For example, to invoke a Function Stage locally in debug mode on port 5858: 
  $ hss invoke -d 5858 <FunctionStage> 
4.4.1  Debugging Python Functions
Python requires you to enable remote debugging in your Function Stage code. If you enable debugging with --debugPort or -d for a Function Stage that uses one of the Python runtimes, HSS will map through that port from your host machine through to the Function Stage runtime container. You will need to enable remote debugging in your function code. To do this, use a python package such as remote-pdb: https://pypi.python.org/pypi/remote-pdb.
4.4.2  Debugging Go Functions
Go debugging using the Delve debugger is not available.


